// IckarSocket.cpp: implementation of the CIckarSocket class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#if defined(WIN32)
#include <winsock2.h>
#include "AsyncProxySocketLayer.h"
#else
#include <errno.h>
#include <arpa/inet.h>
#endif
#include "IckarSocket.h"

#ifndef IP_TOS
#define IP_TOS 8
#endif

#include "CustomDefine.h"

//#include "GTLog.h"

#define MAX_PACKET_BUFFER_COUNT     170

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CIckarSocket::CIckarSocket()
{
	DLOG("CIckarSocket::CIckarSocket(), this=%p", this);
	m_nFiringEventCount = 0;
	m_bDeleteMark = false;
	m_bDisconnected = false;
	m_bNotifyClosed = false;
	m_pEvents = NULL;
	m_nLastError = 0; 
	m_bConnected = false;
#if defined(WIN32)
	m_pProxyLayer = NULL;
#endif
	m_bUseRTSMode = false;
	
	m_mutex.SetRecursive();
	m_mutexEvent.SetRecursive();
}


CIckarSocket::~CIckarSocket()
{
	DLOG("[ENTER] CIckarSocket::~CIckarSocket(), this=%p, m_hSocket=%d %p", this, m_hSocket, GetCurrentThreadId());

	Close();
#if defined(WIN32)
	if( m_pProxyLayer )
		delete m_pProxyLayer;
#endif
	
	DLOG("[LEAVE] CIckarSocket::~CIckarSocket(), this=%p, m_hSocket=%d", this, m_hSocket);
}

#if defined(WIN32)
void CIckarSocket::CheckSocketNotify()
{
	CAsyncSocketEx::CheckOnSendEvent();	
}

bool CIckarSocket::AttachSocket(int socket)
{
	Close();
	if( !Attach( (SOCKET)socket ) )
		return false;

	m_bConnected = true;

	return true;
}

#endif

int CIckarSocket::DetachSocket()
{
	return Detach();
}


int CIckarSocket::GetSocketHandle()
{
	return CAsyncSocketEx::GetSocketHandle();
}


bool CIckarSocket::Bind(int nBindPort, LPCSTR szBindAddr)
{
	/*
	if (m_hSocket != INVALID_SOCKET)
		Close();
	
	if( szBindAddr == NULL || strlen(szBindAddr) == 0 )
	{
		return TRUE == CAsyncSocketEx::Create( nBindPort );
	}

	if ( !CAsyncSocketEx::Socket() )
	{
		m_nLastError = CAsyncSocketEx::GetLastError();
		return false;
	}

	if (!CAsyncSocketEx::Bind(nBindPort, szBindAddr))
	{
		m_nLastError = CAsyncSocketEx::GetLastError();

		CAsyncSocketEx::Close();
		return false;
	}

	return true;
	*/

	return BindEx( nBindPort, szBindAddr, eNCS_PT_TCP );
}

bool CIckarSocket::Connect(LPCSTR szHostAddr, int nHostPort, int nTimeOut /*= -1*/)
{
	return ConnectEx( szHostAddr, nHostPort, eNCS_PT_TCP, nTimeOut );
}

void CIckarSocket::Close()
{
	m_bUseRTSMode = false;

	DLOG( "[ENTER] CIckarSocket::Close(), this=%p, m_hSocket=%d m_queueSendData.size()=%d", this, m_hSocket, m_queueSendData.size() );

	m_bDisconnected = true;
	m_bConnected = false;

	m_mutex.Lock();
	
	while (!m_queueSendData.empty())
	{
		delete m_queueSendData.front();
		m_queueSendData.pop_front();
	}
	
	m_mutex.Unlock();

	CAsyncSocketEx::Close();
#if defined(WIN32)
	if( m_pProxyLayer )
	{
		delete m_pProxyLayer;
		m_pProxyLayer = NULL;
	}
#endif
	DLOG( "[LEAVE] CIckarSocket::Close(), this=%p",this );
}


#if defined(WIN32)
void CIckarSocket::ShutDown(E_NCS_SD_TYPE eType)
{
	CAsyncSocketEx::ShutDown( eType );
}
#endif
bool CIckarSocket::GetSockName(int & nSocketPort, char szSocketAddress[], int nMax)
{
	SOCKADDR_IN sockAddr;
	memset(&sockAddr, 0, sizeof(sockAddr));

	int nSockAddrLen = sizeof(sockAddr);
	if( !CAsyncSocketEx::GetSockName((SOCKADDR*)&sockAddr, &nSockAddrLen) )
		return false;

	nSocketPort = ntohs(sockAddr.sin_port);
	char * pszSockAddr = inet_ntoa(sockAddr.sin_addr);
	if (strlen(pszSockAddr) >= nMax)
		return false;

	if( szSocketAddress )
	{
		if( nMax > strlen(pszSockAddr) )
		{
			strcpy(szSocketAddress, pszSockAddr);
			return true;
		}
		else
		{
			strncpy(szSocketAddress, pszSockAddr, nMax-1);
			szSocketAddress[nMax-1] = 0;
			return false;
		}
	}

	return true;
}

bool CIckarSocket::GetPeerName(int & nPeerPort, char szPeerAddress[], int nMax)
{
	SOCKADDR_IN sockAddr;
	memset(&sockAddr, 0, sizeof(sockAddr));

	int nSockAddrLen = sizeof(sockAddr);
	if( !CAsyncSocketEx::GetPeerName((SOCKADDR*)&sockAddr, &nSockAddrLen) )
		return false;

	nPeerPort = ntohs(sockAddr.sin_port);
	char * pszSockAddr = inet_ntoa(sockAddr.sin_addr);
	if (strlen(pszSockAddr) >= nMax)
		return false;

	if( szPeerAddress )
	{
		if( nMax > strlen(pszSockAddr) )
		{
			strcpy(szPeerAddress, pszSockAddr);
			return true;
		}
		else
		{
			strncpy(szPeerAddress, pszSockAddr, nMax-1);
			szPeerAddress[nMax-1] = 0;
			return false;
		}
	}

	return true;
}

void CIckarSocket::Send(LPCSTR szData, int nLen)
{
//	DLOG( "[ENTER] CIckarSocket::Send( szData=[%s], nLen=%d ), this=%p, m_hSocket=%d", szData, nLen, this, m_hSocket );

#ifdef __USE_BSDSOCKET_WITH_KQUEUE__
	
	m_mutex.Lock();
	
    DLOG ("CIckarSocket::Send : this=%p, Buffering=%d", this, m_queueSendData.size());
    
    if(m_queueSendData.size() > MAX_PACKET_BUFFER_COUNT)
    {
        CAsyncSocketEx::Close();
        OnClose(-1);
        
        m_mutex.Unlock();
        return;
    }
    
	ICKARSOCKET_SEND_QUEUE_DATA * pData = new ICKARSOCKET_SEND_QUEUE_DATA(szData, nLen);
	m_queueSendData.push_back(pData);
	
	m_mutex.Unlock();
	
	//DLOG("CIckarSocket::SEND %d : %d\n", m_hSocket,  m_queueSendData.size());
	CAsyncSocketEx::CheckOnSendEvent();	
	
#if 0	
	if( !m_bUseRTSMode && !m_queueSendData.empty() )
	{
		ASSERT (m_bSending = true);
		ICKARSOCKET_SEND_QUEUE_DATA * pData = new ICKARSOCKET_SEND_QUEUE_DATA(szData, nLen);
		m_queueSendData.push_back(pData);
		return;
	}
	
	int n = CAsyncSocketEx::Send(szData, nLen);
	
	//uint32_t end = GetTickCount();
	//if ( end - start > 0 )
	//	printf("CAsyncSocketEx::Send = %d\n", end - start );
	
	if (n < nLen)
	{
		if (n < 0)
		{			
			m_nLastError = CAsyncSocketEx::GetLastError();
			if (m_nLastError != EWOULDBLOCK)
			{
				OnSocketError( m_nLastError );
				return;
			}				
			n = 0;
		}
		
		if( !m_bUseRTSMode )
		{
			ICKARSOCKET_SEND_QUEUE_DATA * pData = new ICKARSOCKET_SEND_QUEUE_DATA(szData + n, nLen - n);
			m_queueSendData.push_back(pData);
		}
	}
	
	if( !m_bUseRTSMode )
	{
		if(m_queueSendData.empty() == false)
		{
			m_bSending = true;
		
			CAsyncSocketEx::CheckOnSendEvent();	
		}
	}
#endif	
	
#else	// __USE_BSDSOCKET_WITH_KQUEUE__
	
	if( !m_bUseRTSMode && !m_queueSendData.empty() )
	{
		ASSERT (m_bSending = true);
		ICKARSOCKET_SEND_QUEUE_DATA * pData = new ICKARSOCKET_SEND_QUEUE_DATA(szData, nLen);
		m_queueSendData.push_back(pData);
		return;
	}
	
	//uint32_t start = GetTickCount();
	
	int n = CAsyncSocketEx::Send(szData, nLen);
	
	//uint32_t end = GetTickCount();
	//if ( end - start > 0 )
	//	printf("CAsyncSocketEx::Send = %d\n", end - start );

#if defined(WIN32)	// __KEH_USE_MAC_SOCKET__
	if (n < nLen)
	{
		if (n < 0)
		{			
			m_nLastError = CAsyncSocketEx::GetLastError();
			if (m_nLastError != WSAEWOULDBLOCK)
			{
				OnSocketError( m_nLastError );
				return;
			}				
			n = 0;
		}

		if( !m_bUseRTSMode )
		{
			ICKARSOCKET_SEND_QUEUE_DATA * pData = new ICKARSOCKET_SEND_QUEUE_DATA(szData + n, nLen - n);
			m_queueSendData.push_back(pData);
		}
	}
#else	// defined(WIN32)
	// Don't need to resend-logic..
	// Already implemented in AsyncSocket.mm
	// If failed to write with sock error, onClose func be called by AsyncSocket.mm
#endif	// defined(WIN32)
	
	if( !m_bUseRTSMode )
	{
		m_bSending = true;
		
		CAsyncSocketEx::CheckOnSendEvent();	
	}

#endif	// __USE_BSDSOCKET_WITH_KQUEUE__
//	DLOG ("[LEAVE] CIckarSocket::Send(), this=%p, m_hSocket=%d", this, m_hSocket);
}

#if defined(WIN32)
void CIckarSocket::Sendv(char* szDataBuff[], USHORT usDataLen[], USHORT count)
{
//	DLOG( "[ENTER] CIckarSocket::Sendv( szData=[%s], nLen=%d ), this=%p, m_hSocket=%d", szData, nLen, this, m_hSocket );

	if( !m_bUseRTSMode && !m_queueSendData.empty() )
	{
		ASSERT (m_bSending = true);
		ICKARSOCKET_SEND_QUEUE_DATA * pData = new ICKARSOCKET_SEND_QUEUE_DATA(szDataBuff, usDataLen, count);
		m_queueSendData.push_back(pData);
		return;
	}

	ULONG ulTotalLen = 0;
	WSABUF wsaBuf[2];
	WSABUF * pWsaBuf = wsaBuf;
	if( count > 2 )
		pWsaBuf = new WSABUF[count];

	for( int i=0 ; i < count ; i++ )
	{
		pWsaBuf[i].buf = szDataBuff[i];
		pWsaBuf[i].len = usDataLen[i];

		ulTotalLen += usDataLen[i];
	}

	DWORD dwNumberOfBytesSent = 0;
	int n = CAsyncSocketEx::SendWSA(pWsaBuf, count, &dwNumberOfBytesSent, 0, NULL, NULL );
	if( count > 2 )
		delete [] pWsaBuf;

	if( n != 0 )	// == SOCKET_ERROR
	{
		m_nLastError = CAsyncSocketEx::GetLastError();
		if (m_nLastError != WSAEWOULDBLOCK)
		{
			OnSocketError( m_nLastError );
			return;
		}
		else
		{
			ASSERT( dwNumberOfBytesSent == 0 );

			if( !m_bUseRTSMode )
			{
				ICKARSOCKET_SEND_QUEUE_DATA * pData = new ICKARSOCKET_SEND_QUEUE_DATA(szDataBuff, usDataLen, count);
				m_queueSendData.push_back(pData);
			}
		}
	}
	else
	{
		ASSERT( dwNumberOfBytesSent == ulTotalLen );
	}
	
	if( !m_bUseRTSMode )
	{
		m_bSending = true;
		CAsyncSocketEx::CheckOnSendEvent();	
	}

//	DLOG ("[LEAVE] CIckarSocket::Sendv(), this=%p, m_hSocket=%d", this, m_hSocket);
}
#endif
 
int CIckarSocket::Receive(LPSTR szData, int nLen)
{
	return CAsyncSocketEx::Receive( szData, nLen);
	/*
	int n = CAsyncSocketEx::Receive( szData, nLen);
	if( n==0 )
	{
		// socket read closed
		OnClose( CAsyncSocketEx::GetLastError() );
		return 0;
	}
	else if( n < 0 )
	{
		m_nLastError = CAsyncSocketEx::GetLastError();
		if (m_nLastError == WSAEWOULDBLOCK)
			return 0;

		OnSocketError( m_nLastError );
		return n;
	}

	return n;
	*/
}

bool CIckarSocket::Listen(int nConnectionBacklog)
{
#if defined(WIN32) || defined(__USE_BSDSOCKET_WITH_KQUEUE__)
	DLOG ("[ENTER] CIckarSocket::Listen(nConnectionBacklog=%d), this=%p", nConnectionBacklog, this);
	if (m_hSocket == INVALID_SOCKET)
		return false;

	DLOG ("[LEAVE] CIckarSocket::Listen(nConnectionBacklog=%d), this=%p", nConnectionBacklog, this);
	return CAsyncSocketEx::Listen(nConnectionBacklog) == TRUE;
#else
	return false;
#endif
}

bool CIckarSocket::Accept(IIckarSocket * pIckarSocket)
{

	DLOG ("[ENTER] CIckarSocket::Accept(pIckarSocket=%p), this=%p", pIckarSocket, this);

	CIckarSocket * pIckarSocketImpl = dynamic_cast<CIckarSocket *>(pIckarSocket);
	if (!pIckarSocketImpl)
		return false;

	CAsyncSocketEx* pAsyncSocketEx = static_cast<CAsyncSocketEx*>(pIckarSocketImpl);
	if (!pAsyncSocketEx)
		return false;
	
#if defined(WIN32)
	SOCKADDR sockaddr;
	int sockaddr_len = sizeof(sockaddr);

	if( !CAsyncSocketEx::Accept( *pAsyncSocketEx, &sockaddr, &sockaddr_len ) )
	{
		m_nLastError = CAsyncSocketEx::GetLastError();

		DLOG ("[LEAVE] CIckarSocket::Accept(pIckarSocket=%p), this=%p, Accept FAIL, m_nLastError=%d", pIckarSocket, this, m_nLastError);
		return false;
	}
	
	pIckarSocketImpl->m_bConnected = true;
	pIckarSocketImpl->AppropriateSocketOption();

	DLOG ("[LEAVE] CIckarSocket::Accept(pIckarSocketImpl=%p, pIckarSocketImpl->m_hSocket=%d), this=%p, m_hSocket=%d", 
		pIckarSocketImpl, pIckarSocketImpl->m_hSocket, this, m_hSocket);

	return pAsyncSocketEx->AsyncSelect( FD_READ | FD_WRITE | FD_CLOSE ) == TRUE;
#else
	//KEH_MOBILE_MARK_P2P
	pIckarSocketImpl->m_bConnected = true;
	return CAsyncSocketEx::Accept( *pAsyncSocketEx );
#endif
}

bool CIckarSocket::IsConnected()
{
	return m_bConnected;
}

void CIckarSocket::Delete()
{
	DLOG("[ENTER] CIckarSocket::Delete() %p, count = %d, thread = %p", this, m_nFiringEventCount, GetCurrentThreadId());

    m_mutexEvent.Lock();

	m_pEvents = NULL;
	
    m_mutexEvent.Unlock();

	if( m_nFiringEventCount == 0 )
	{
		CAsyncSocketEx::Delete();
	}
	else
		m_bDeleteMark = true;
	DLOG("[LEAVE] CIckarSocket::Delete() %p %d, thread = %p", this, m_bDeleteMark, GetCurrentThreadId());
}

int CIckarSocket::ReceiveFrom(LPSTR lpBuf, int nBufLen,
		USHORT &uRemotePort, char szHostIP[], int nMax /*= 16*/, int nFlags /*= 0*/)
{
	SOCKADDR_IN sockAddr;
	memset(&sockAddr, 0, sizeof(sockAddr));

	socklen_t nSockAddrLen = sizeof(sockAddr);
	int n = CAsyncSocketEx::ReceiveFrom( lpBuf, nBufLen, (SOCKADDR*)&sockAddr, &nSockAddrLen, nFlags );
	if( n < 0 )
	{
		m_nLastError = CAsyncSocketEx::GetLastError();
		return n;
	}

	uRemotePort = ntohs(sockAddr.sin_port);
	char * pszSockAddr = inet_ntoa(sockAddr.sin_addr);
	if (strlen(pszSockAddr) >= nMax)
	{
		ASSERT("CIckarSocket::ReceiveFrom, ERROR, buffer length too short!!" && false );
		strncpy(szHostIP, pszSockAddr, nMax);
		szHostIP[nMax] = 0;
	}
	else
	{
		strcpy(szHostIP, pszSockAddr);
	}

	return n;
}

int CIckarSocket::SendTo(LPCSTR szData, int nDataLen,
		USHORT uHostPort, LPCSTR szHostIP, int nFlags /*= 0*/)
{
	SOCKADDR_IN sockAddr;
	memset(&sockAddr, 0, sizeof(sockAddr));

	sockAddr.sin_family=AF_INET;
	sockAddr.sin_addr.s_addr = inet_addr( szHostIP );		
	sockAddr.sin_port = htons(uHostPort);

	return CAsyncSocketEx::SendTo( szData, nDataLen, (SOCKADDR*)&sockAddr, sizeof(sockAddr), nFlags );
}

#if defined(WIN32)
bool CIckarSocket::SetRTSMode()
{
	m_bUseRTSMode = true;

	int usTOS = 0xB8;
	int len = sizeof(usTOS);

	int result;
	result = getsockopt( m_hSocket, IPPROTO_IP, IP_TOS, (char *)&usTOS, &len ); 
	if( result == SOCKET_ERROR )
	{
		m_nLastError = CAsyncSocketEx::GetLastError();
	}

	usTOS = 0xB8;
	len = sizeof(usTOS);
	result = setsockopt( m_hSocket, IPPROTO_IP, IP_TOS, (char *)&usTOS, len ); 
	if( result == SOCKET_ERROR )
	{
		m_nLastError = CAsyncSocketEx::GetLastError();
	}
	
	len = sizeof(usTOS);
	result = getsockopt( m_hSocket, IPPROTO_IP, IP_TOS, (char *)&usTOS, &len ); 
	if( result == SOCKET_ERROR )
	{
		m_nLastError = CAsyncSocketEx::GetLastError();
	}

	CAsyncSocketEx::AsyncSelect( 0 );

	return ( result != SOCKET_ERROR );
}
#endif

bool CIckarSocket::ConnectEx(LPCSTR szHostAddr, int nHostPort, E_NCS_PROTOCOL_TYPE ePType /*= eNCS_PT_TCP*/, int nTimeOut /*= -1*/)
{
	DLOG("[ENTER] CIckarSocket::ConnectEx( szHostAddr=%s, nHostPort=%d, ePType=%d ), this=%p", szHostAddr, nHostPort, ePType, this );

	ASSERT( !IsConnected() );
	if( IsConnected() )
	{
//		GT_FLOG(L"Already socket is connected");
		Close();
		return false;
	}

	bool bRes = true;

	if( m_hSocket == INVALID_SOCKET )
	{
		if( !CAsyncSocketEx::Create( 0, ePType == eNCS_PT_TCP ? SOCK_STREAM : SOCK_DGRAM ) )//,
//				ePType == eNCS_PT_UDP ? FD_WRITE | FD_CLOSE : FD_READ | FD_WRITE | FD_OOB | FD_ACCEPT | FD_CONNECT | FD_CLOSE ) )
		{
			bRes = false;
			m_nLastError = CAsyncSocketEx::GetLastError();
//			GT_FLOG(L"Fail to create socket");
		}
	}
	
	if( bRes )
	{
		AppropriateSocketOption();

		if (CAsyncSocketEx::Connect(szHostAddr, nHostPort))
		{
			m_bConnected = true;
			bRes = true;
		}
		else
		{
			m_nLastError = CAsyncSocketEx::GetLastError();
			if (m_nLastError == WSAEWOULDBLOCK)
			{
				bRes = true;
			}
			else
			{
				//GT_FLOG(L"Fail to connect to server [%s], [%d] : [%d]", szHostAddr, nHostPort, m_nLastError);
				bRes = false;
			}
		}
	}

	DLOG("[LEAVE] CIckarSocket::ConnectEx( szHostAddr=%s, nHostPort=%d, ePType=%d ), this=%p, m_hSocket=%d\n"
		"m_nLastError=%d\n"
		"bRes=%d", szHostAddr, nHostPort, ePType, this, m_hSocket, m_nLastError, bRes );

	//GT_FLOG(L"szHostAddr=%s, nHostPort=%d, ePType=%d, this=%p, m_hSocket=%d, m_nLastError=%d, bRes=%d", szHostAddr, nHostPort, ePType, this, m_hSocket, m_nLastError, bRes );

	return bRes;
}

bool CIckarSocket::BindEx(int nBindPort, LPCSTR szBindAddr, E_NCS_PROTOCOL_TYPE ePType /*= eNCS_PT_TCP*/)
{
#if defined(WIN32)	
	if( m_hSocket != INVALID_SOCKET )
		Close();
	
	if( szBindAddr == NULL || strlen(szBindAddr) == 0 )
	{
		return TRUE == CAsyncSocketEx::Create( nBindPort, ePType == eNCS_PT_TCP ? SOCK_STREAM : SOCK_DGRAM);//, 
//			ePType == eNCS_PT_UDP ? FD_WRITE | FD_CLOSE : FD_READ | FD_WRITE | FD_OOB | FD_ACCEPT | FD_CONNECT | FD_CLOSE );
	}
	
	if( !CAsyncSocketEx::Socket(ePType == eNCS_PT_TCP ? SOCK_STREAM : SOCK_DGRAM) )//,
//		ePType == eNCS_PT_UDP ? FD_WRITE | FD_CLOSE : FD_READ | FD_WRITE | FD_OOB | FD_ACCEPT | FD_CONNECT | FD_CLOSE ) )
	{
		m_nLastError = CAsyncSocketEx::GetLastError();
		return false;
	}

#if defined(WIN32)
	if( !CAsyncSocketEx::Bind(nBindPort, szBindAddr) )
#else
	if( !CAsyncSocketEx::Bind((unsigned int)nBindPort, (const wchar_t*)szBindAddr) )
#endif
	{
		m_nLastError = CAsyncSocketEx::GetLastError();

		CAsyncSocketEx::Close();
		return false;
	}

	return true;
#else
	
#ifdef __USE_BSDSOCKET_WITH_KQUEUE__	
	if( m_hSocket != INVALID_SOCKET )
		Close();
	
	CAsyncSocketEx::Create( nBindPort, ePType == eNCS_PT_TCP ? SOCK_STREAM : SOCK_DGRAM);
#endif
	
	if(!CAsyncSocketEx::Bind(nBindPort))
	{
		// MUST NOT CLOSE SOCKET!!
		//CAsyncSocketEx::Close();
		return false;
	}
	
	return true;
#endif
}


#if defined(WIN32)
E_NCS_PROTOCOL_TYPE CIckarSocket::GetNcsProtocalType()
{
	WSAPROTOCOL_INFO wsaProtocal;
	int len = sizeof(WSAPROTOCOL_INFO);
	ZeroMemory( &wsaProtocal, len );
	
	GetSockOpt( SO_PROTOCOL_INFO, &wsaProtocal, &len );
	
	return (wsaProtocal.iSocketType == SOCK_STREAM) ? eNCS_PT_TCP : eNCS_PT_UDP;
}
#endif
////////////////////////////////////////////




#if !defined(WIN32)
#pragma mark -
#pragma mark CAsyncSocketEx Event
#endif

void CIckarSocket::OnSocketError(int nErrorCode)
{
	DLOG ("[ENTER] CIckarSocket::OnSocketError(nErrorCode=%d), this=%p, m_hSocket=%d", nErrorCode, this, m_hSocket);
	m_nLastError = nErrorCode;
#if defined(WIN32)
	CAsyncSocketEx::TriggerEvent(FD_CLOSE, nErrorCode);
#else
	OnClose(nErrorCode);
#endif
	
	DLOG ("[LEAVE] CIckarSocket::OnSocketError(nErrorCode=%d), this=%p", nErrorCode, this);
}

void CIckarSocket::OnReceive(int nErrorCode, int nBytes)
{
	if (nErrorCode)
	{
		OnClose(nErrorCode);
		return;
	}
	
	if(nBytes < 0)
	{
#if defined(WIN32)
		if (!IOCtl(FIONREAD, (DWORD*)&nBytes))
		{
			nErrorCode = CAsyncSocketEx::GetLastError();

			OnClose(nErrorCode);
			return;
		}
#endif
	}

	InterlockedIncrement(&m_nFiringEventCount);
	// KEH_API_STABLE_PROJECT
	// NULL 체크
    
    m_mutexEvent.Lock();

	if(m_pEvents)
		m_pEvents->OnIckarSocket_ReceiveReady( this, nBytes );
    else
    {
        // ERROR..
		DLOG(">> OnReceive : m_pEvents is NULL..");
    }

    m_mutexEvent.Unlock();

	InterlockedDecrement(&m_nFiringEventCount);
	
	if( Check_AsyncDelete() )
		return;
}


void CIckarSocket::OnSend(int nErrorCode)
{
	if (nErrorCode)
	{
		OnClose(nErrorCode);
		return;
	}

	if( !m_queueSendData.empty() )
	{
		int nHandleCount = 0;
		do
		{
			m_mutex.Lock();
			
			if(m_queueSendData.empty())
			{
				DLOG("ALREADY SEND QUEUE EMPTY ON OTHER THREAD!!\n");
				m_mutex.Unlock();
				return;
			}
			
			ICKARSOCKET_SEND_QUEUE_DATA * pData = m_queueSendData.front();

			int n = CAsyncSocketEx::Send( pData->m_pData, pData->m_nLen );

			if( n == pData->m_nLen )
			{
				nHandleCount++;
				m_queueSendData.pop_front();
				
				delete pData;
			}
			else
			{
				if (n <= 0)		// socket error
				{
					if (!GetLastWouldBlockError())
					{
						OnSocketError( m_nLastError );
					}
				}
				else
				{
					m_queueSendData.pop_front();

					ICKARSOCKET_SEND_QUEUE_DATA * pData2 = new ICKARSOCKET_SEND_QUEUE_DATA(pData->m_pData + n, pData->m_nLen - n);
					m_queueSendData.push_front(pData2);

					delete pData;
				}

				m_mutex.Unlock();
				return;
			}
			
			m_mutex.Unlock();
		}
		while( !m_queueSendData.empty() );
		
		//DLOG("ONSEND %d : %d(%d)\n", m_hSocket, m_queueSendData.size(), nHandleCount);
		
		m_mutex.Lock();
		
		if(m_queueSendData.empty())
			CAsyncSocketEx::UnCheckOnSendEvent();
		else
			CAsyncSocketEx::CheckOnSendEvent();

		m_mutex.Unlock();
		return;
	}

	if ( m_bSending )
	{
		m_bSending = false;

		InterlockedIncrement(&m_nFiringEventCount);
		// KEH_API_STABLE_PROJECT
		// NULL 체크
        m_mutexEvent.Lock();

		if(m_pEvents)
			m_pEvents->OnIckarSocket_SendReady(this);
        
        m_mutexEvent.Unlock();

		InterlockedDecrement(&m_nFiringEventCount);
		
		if( Check_AsyncDelete() )
			return;
	}
}

void CIckarSocket::OnConnect(int nErrorCode)
{
	DLOG("[ENTER] CIckarSocket::OnConnect(nErrorCode=%d), this=%p, m_hSocket=%d", nErrorCode, this, m_hSocket);

	if(m_bRequestConnect == false)  // __KEH_USE_MAC_SOCKET__
		return;
    m_bRequestConnect = false;
    
	if (nErrorCode == 0)
		m_bConnected = true;

	m_nLastError = nErrorCode;

	m_bSending = true;
	InterlockedIncrement(&m_nFiringEventCount);

	// KEH_API_STABLE_PROJECT
	// NULL 체크
    m_mutexEvent.Lock();

	if(m_pEvents)
		m_pEvents->OnIckarSocket_Connect( this, nErrorCode );
    
    m_mutexEvent.Unlock();

	InterlockedDecrement(&m_nFiringEventCount);
	
	if( Check_AsyncDelete() )
		return;

	DLOG("[LEAVE] CIckarSocket::OnConnect(nErrorCode=%d), this=%p, m_hSocket=%d", nErrorCode, this, m_hSocket);
}

void CIckarSocket::OnAccept(int nErrorCode)
{
	DLOG("[ENTER] CIckarSocket::OnAccept(nErrorCode=%d), this=%p, m_hSocket=%d", nErrorCode, this, m_hSocket);

	if (nErrorCode)
	{
		OnClose(nErrorCode);
		return;
	}

	InterlockedIncrement(&m_nFiringEventCount);

	// KEH_API_STABLE_PROJECT
	// NULL 체크
    m_mutexEvent.Lock();

	if(m_pEvents)
		m_pEvents->OnIckarSocket_Accept( this, nErrorCode );
    
    m_mutexEvent.Unlock();

	InterlockedDecrement(&m_nFiringEventCount);
	
	if( Check_AsyncDelete() )
		return;

	DLOG("[LEAVE] CIckarSocket::OnAccept(nErrorCode=%d), this=%p, m_hSocket=%d", nErrorCode, this, m_hSocket);
}

void CIckarSocket::OnClose(int nErrorCode)
{
	DLOG("[ENTER] CIckarSocket::OnClose(nErrorCode=%d), this=%p, m_hSocket=%d", nErrorCode, this, m_hSocket);

	if( m_bNotifyClosed )
	{
#if defined(WIN32)
#ifdef _DEBUG
		char szMsg[1024];
		_snprintf(szMsg, 1024, "CIckarSocket::OnClose, pSocket->m_bNotifyClosed==true!!, m_hSocket=%d, nErrorCode=%d, m_nSMT_Index=%d",
			m_hSocket,
			nErrorCode,
			m_nSMT_Index );

		//MessageBox(NULL, szMsg, "[CIkcarSocket]", MB_OK|MB_ICONERROR);
#endif
#endif
		return;
	}
	
	m_bDisconnected = true;
	m_bConnected = false;
	m_nLastError = nErrorCode;

#if defined(WIN32)
	SOCKET hSocket_backup = m_hSocket;
#endif
	m_bNotifyClosed = true;

	InterlockedIncrement(&m_nFiringEventCount);

	// KEH_API_STABLE_PROJECT
	// NULL 체크
    m_mutexEvent.Lock();

	if(m_pEvents)
	{
		m_pEvents->OnIckarSocket_Close( this, nErrorCode );
	}
	else 
	{
		// ERROR..
		DLOG(">> m_pEvents is NULL..");
	}
    m_mutexEvent.Unlock();

	InterlockedDecrement(&m_nFiringEventCount);
	
	if( Check_AsyncDelete() )
		return;
#if defined(WIN32)
	DLOG("[LEAVE] CIckarSocket::OnClose(nErrorCode=%d), this=%p, hSocket_backup=%d, m_hSocket=%d", nErrorCode, this, hSocket_backup, m_hSocket);
#endif
}


